package structures.map;

import procedures.combinations.ListCons;
import procedures.conversions.EntryKey;
import procedures.conversions.EntryValue;
import procedures.orderings.EntryOrder;
import procedures.predicates.AVLInvariant;
import algorithms.ListAlgorithms;
import algorithms.TreeAlgorithms;
import structures.list.Empty;
import structures.tree.Leaf;
import types.procedures.ICompare;
import types.procedures.ISelect;
import types.structures.IMap;
import types.structures.ISame;
import types.structures.ITree;
import types.structures.IList;
import exceptions.NotFoundException;

// Represents a mapping with ordered keys.
public class OrderedMap implements IMap {

	private static ISelect balanceInvariant = new AVLInvariant();
	
	public ICompare keyOrder;
	private ICompare entryOrder;
	private ITree tree;
	
	public OrderedMap( ICompare keyOrder ) {
		this.keyOrder = keyOrder;
		this.entryOrder = new EntryOrder(keyOrder);
		this.tree = new Leaf();
	}
	
	private OrderedMap( OrderedMap map, ITree tree ) {
		this.keyOrder = map.keyOrder;
		this.entryOrder = map.entryOrder;
		this.tree = tree;
	}

	// Produces the Entry corresponding to key, or throws NotFoundException.
	private Entry find( ISame key ) {
		return (Entry)TreeAlgorithms.find( this.tree, new Entry( key, null), entryOrder );
	}

	// Tells whether this has no key/value mappings.
	public boolean isEmpty() {
		return tree.isLeaf();
	}
	
	// Tells whether this maps key to a value.
	public boolean contains(ISame key) {
		try {
			this.find( key );
			return true;
		} catch( NotFoundException e ) {
			return false;
		}
	}

	// Produces the value which key is mapped to, or throws NotFoundException.
	public Object lookup(ISame key) {
		return this.find(key).value;
	}

	// Produces a copy of this, mapping key to value.
	public IMap extend(ISame key, Object value) {
		ITree tree =
			TreeAlgorithms.insertBalanced(
					this.tree, new Entry( key, value ), entryOrder, balanceInvariant );
		return new OrderedMap( this, tree );
	}

	// Produces a list of all keys, sorted by the given ordering.
	public IList keys() {
		return ListAlgorithms.map( this.entries(), new EntryKey() );
	}

	// Produces a list of all values, sorted to correspond to the keys.
	public IList values() {
		return ListAlgorithms.map( this.entries(), new EntryValue() );
	}

	// Produces a list of all key/value pairs, sorted by the ordering on keys.
	public IList entries() {
		return (IList)TreeAlgorithms.foldRight( this.tree, new Empty(), new ListCons() );
	}
	
	// Produces a string to display this.
	public String toString() {
		return "(map " + this.entries().toString() + ")";
	}

}
